/*-
* Copyright (C) 2005-2010 Nick Withers. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) AND CONTRIBUTORS ``AS IS'' AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR(S) OR CONTRIBUTORS BE LIABLE
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
* OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
* LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
* OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGE.
*/

#include "downtime.h"
#include "interface.h"

#include <errno.h>
#include <gtk/gtk.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <sys/param.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

#ifdef BSD
#include <sysexits.h>
#endif

#include <time.h>

typedef struct {
	GtkWidget *timeout_field;
	GtkWidget *timeout_unit_combo_box;
	GtkWidget *shutdown_type_combo_box;
} ui_specs_elements_t;

typedef struct {
	GtkWidget *main_window_on_top_box;
	GtkWidget *progress_window_on_top_box;
	GtkWidget *show_tray_box;
} ui_prefs_elements_t;

static void destroy_progress_cb(void ***args);
static void destroy_progress(shutdown_t *shutdown, GtkWidget *progress_window);
static void exit_and_abort(shutdown_t *shutdown);
static void exit_no_abort(void);
static void pause_shutdown_cb(void ***args);
static void pause_shutdown(shutdown_t *shutdown, GtkWidget *progress_bar, GtkStatusIcon *tray_icon);
static double progress_fraction(shutdown_t *shutdown);
static void save_as_cb(void ***args);
static void save_as(char **config_path, user_specs_t *specs, ui_specs_elements_t *ui_specs_elements, GtkWidget *main_window);
static void save_current_cb(void ***args);
static void save_current(user_specs_t *specs, char *config_path, ui_specs_elements_t *ui_specs_elements);
static void save_prefs_cb(void ***args);
static void save_prefs(user_specs_t *specs, char *config_path, GtkWidget *window, ui_prefs_elements_t *ui_prefs_elements, GtkWidget *main_window);
static void select_and_open_config_cb(void ***);
static void select_and_open_config(user_specs_t *specs, char **config_path, GtkWidget *main_window, ui_specs_elements_t *ui_specs_elements, GtkWidget *save_menu_item);
static void show_about(void);
static void show_prefs_cb(void ***args);
static void show_prefs(user_specs_t *specs, char *config_path, GtkWidget *main_window);
static GtkStatusIcon *show_tray_icon(GtkWidget *show_hide_window, void ***destroy_progress_cb_args);
static void tray_icon_click_cb(void ***args);
static void tray_icon_click(GtkWidget *show_hide_window);
static void tray_icon_menu(GtkStatusIcon *tray_icon, guint button, guint32 activate_time, gpointer menu);
static shutdown_t * ui_start_shutdown_cb(void ***args);
static shutdown_t * ui_start_shutdown(user_specs_t *specs, GtkWidget *main_window, ui_specs_elements_t *ui_specs_elements);
static void update_main(user_specs_t *specs, char *config_path, ui_specs_elements_t *ui_specs_elements, GtkWidget *save_menu_item);
static gboolean update_progress_cb(void ***args);
static gboolean update_progress(shutdown_t *shutdown, GtkWidget *progress_bar, GtkStatusIcon *tray_icon, GtkWidget *pause_button);
static user_specs_t * update_user_specs(user_specs_t *specs, ui_specs_elements_t *ui_specs_elements);

static char no_check_proc = FALSE;

static void
destroy_progress_cb(void ***args)
{
	destroy_progress(*args[0], *args[1]);
}

static void
destroy_progress(shutdown_t *shutdown, GtkWidget *progress_window)
{
	if (shutdown->paused)
		gtk_main_quit();
	else if (shutdown->pid)
	{
		GtkWidget *window, *vbox, *label, *hbox, *button;
		char *shutdown_pending_message;

		window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
		gtk_window_set_transient_for(GTK_WINDOW (window), GTK_WINDOW (progress_window));
		gtk_window_set_modal(GTK_WINDOW (window), TRUE);
		gtk_window_set_title(GTK_WINDOW (window), "Abort active shutdown? - " PRGNAME);
		gtk_window_set_resizable(GTK_WINDOW (window), FALSE);
		gtk_window_set_position(GTK_WINDOW (window), GTK_WIN_POS_CENTER);
		gtk_container_set_border_width (GTK_CONTAINER (window), 10);
		g_signal_connect(G_OBJECT (window), "delete_event", G_CALLBACK (gtk_widget_destroy), NULL);
		g_signal_connect(G_OBJECT (window), "destroy", G_CALLBACK (gtk_main_quit), NULL);

		vbox = gtk_vbox_new(FALSE, 0);

		if ((pasprintf(&shutdown_pending_message, "There is a shutdown pending (PID %d), would you like to terminate it?", shutdown->pid)) == -1)
			show_error_dialog(EX_OSERR, "pasprintf() returned -1 in %s(). Error: \"%s\" (%d)", __func__, strerror(errno), errno);

		label = gtk_label_new(shutdown_pending_message);
		gtk_box_pack_start(GTK_BOX (vbox), label, FALSE, FALSE, 5);
		gtk_widget_show(label);

		free(shutdown_pending_message);	/* No longer required by the GTK, it stores the string itself */

		hbox = gtk_hbox_new(FALSE, 0);

		gtk_box_pack_start(GTK_BOX (vbox), hbox, FALSE, FALSE, 5);

		button = gtk_button_new_from_stock(GTK_STOCK_YES);
		g_signal_connect_swapped(G_OBJECT (button), "clicked", G_CALLBACK (exit_and_abort), shutdown);
		gtk_box_pack_start(GTK_BOX (hbox), button, FALSE, FALSE, 5);
		gtk_widget_show(button);

		button = gtk_button_new_from_stock(GTK_STOCK_NO);
		g_signal_connect_swapped(G_OBJECT (button), "clicked", G_CALLBACK (exit_no_abort), NULL);
		gtk_box_pack_start(GTK_BOX (hbox), button, FALSE, FALSE, 5);
		gtk_widget_show(button);

		button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
		g_signal_connect_swapped(G_OBJECT (button), "clicked", G_CALLBACK (gtk_widget_destroy), window);
		gtk_box_pack_start(GTK_BOX (hbox), button, FALSE, FALSE, 5);
		gtk_widget_show(button);

		gtk_container_add(GTK_CONTAINER (window), vbox);

		gtk_widget_show(hbox);
		gtk_widget_show(vbox);
		gtk_widget_show(window);

		gtk_main();
	}
	else
	{
		show_warning_dialog("Shutdown PID not known, please terminate the relevant shutdown process manually, if so desired");

		gtk_main_quit();
	}
}

static void
exit_and_abort(shutdown_t *shutdown)
{
	if (kill(shutdown->pid, SIGTERM) != 0)	/* shutdown->pid is checked to ensure it's > 0 in shutdown.c */
		show_error_dialog(EX_OSERR, "Couldn't kill PID %d, error: \"%s\" (%d).\n\nPlease kill the shutdown(8) process manually.", shutdown->pid, strerror(errno), errno);

	exit(EXIT_SUCCESS);
}

static void
exit_no_abort(void)
{
	exit(EXIT_SUCCESS);
}

static void
pause_shutdown_cb(void ***args)
{
	pause_shutdown(*args[0], *args[1], *args[2]);
}

static void
pause_shutdown(shutdown_t *shutdown, GtkWidget *progress_bar, GtkStatusIcon *tray_icon)
{
	if (!shutdown->paused)	/* The shutdown isn't currently paused */
	{
		time_t time_left;

		shutdown->paused = TRUE;	/* Indicate that a shutdown is paused. This'll stop the progress bar from updating */

		if (kill(shutdown->pid, SIGTERM) != 0)	/* We're not checking that the PID is valid here as this has already been done in start_shutdown(), which must have been successfully called if we're here */
			show_warning_dialog("Couldn't kill PID %d in order to pause the shutdown. Error: \"%s\" (%d)", shutdown->pid, strerror(errno), errno);

		time_left = max_minutes_to_shutdown(shutdown) * 60;

		shutdown->start_time += time_left - shutdown->shutdown_time;	/* shutdown->start_time now holds the start time relative to the current time (<= 0, rounded given the granularity of the timer), rather than relative to the epoch */
		shutdown->shutdown_time = time_left;	/* shutdown->shutdown_time now holds the time to go when we resume */

		gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR (progress_bar), progress_fraction(shutdown));
		gtk_progress_bar_set_text(GTK_PROGRESS_BAR (progress_bar), "(Paused)");

		if (tray_icon)
			gtk_status_icon_set_tooltip(tray_icon, "DownTime - (Paused)");
	}
	else
	{
		if ((start_shutdown(shutdown)) == NULL)
			show_error_dialog(EX_OSERR, "Couldn't re-launch shutdown(8) following a pause. Error: \"%s\" (%d)", strerror(errno), errno);
	}
}

static double
progress_fraction(shutdown_t *shutdown)
{
	double new_progress;

	new_progress = (shutdown->paused ? -shutdown->start_time : difftime(time(NULL), shutdown->start_time)) / difftime(shutdown->shutdown_time, shutdown->start_time);

	if (new_progress > 1)
		new_progress = 1;
	else if (new_progress < 0)
		new_progress = 0;

	return (new_progress);
}

static void
save_as_cb(void ***args)
{
	save_as(*args[0], *args[1], *args[2], *args[3]);
}

static void
save_as(char **config_path, user_specs_t *specs, ui_specs_elements_t *ui_specs_elements, GtkWidget *main_window)
{
	GtkWidget *chooser, *auto_run_box;
	GtkFileFilter *config_filter;

	chooser = gtk_file_chooser_dialog_new("Save Configuration File", GTK_WINDOW (main_window), GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);

	gtk_file_chooser_set_do_overwrite_confirmation(GTK_FILE_CHOOSER (chooser), TRUE);
	auto_run_box = gtk_check_button_new_with_label("Auto-start shutdown when this configuration is launched");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (auto_run_box), TRUE);
	gtk_widget_show(auto_run_box);

/* "If you want to set more than one extra widget in the file chooser, you can a container such as a GtkVBox or a GtkTable and include your widgets in it. Then, set the container as the whole extra widget." */
	gtk_file_chooser_set_extra_widget(GTK_FILE_CHOOSER (chooser), auto_run_box);

	gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER (chooser), "Untitled." CONFIG_EXTENSION);

	config_filter = gtk_file_filter_new();
	gtk_file_filter_add_pattern(config_filter, "*." CONFIG_EXTENSION);
	gtk_file_filter_add_mime_type(config_filter, CONFIG_MIME_TYPE);
	gtk_file_filter_set_name(config_filter, PRGNAME " configuration");
	gtk_file_chooser_set_filter(GTK_FILE_CHOOSER (chooser), config_filter);

	if (gtk_dialog_run(GTK_DIALOG (chooser)) == GTK_RESPONSE_ACCEPT)
	{
		char *filename;

		filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER (chooser));

		{
			char *temp;

			if ((pasprintf(&temp, "%s", filename)) == -1)
				show_error_dialog(EX_OSERR, "pasprintf() returned -1 in %s(). Error: \"%s\" (%d)", __func__, strerror(errno), errno);

			g_free(filename);

			filename = temp;
		}

		specs->auto_run = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (auto_run_box));

		if (update_user_specs(specs, ui_specs_elements) == NULL)
		{
			show_warning_dialog("Could not save settings. Error: \"%s\" (%d)", strerror(errno), errno);

			return;
		}

		if (save(specs, filename) == FALSE)
			show_warning_dialog("Unable to save to configuration file \"%s\".\n\nError \"%s\" (%d)", filename, strerror(errno), errno);

		free(*config_path);
		*config_path = filename;
	}

	gtk_widget_destroy(chooser);
}

static void
save_current_cb(void ***args)
{
	save_current(*args[0], *args[1], *args[2]);
}

static void
save_current(user_specs_t *specs, char *config_path, ui_specs_elements_t *ui_specs_elements)
{
	if (update_user_specs(specs, ui_specs_elements) == NULL)
		return;

	if (save(specs, config_path) == FALSE)
		show_warning_dialog("Unable to save to configuration file \"%s\". Settings not saved\n\nError \"%s\" (%d)", config_path, strerror(errno), errno);
}

static void
save_prefs_cb(void ***args)
{
	save_prefs(*args[0], *args[1], *args[2], *args[3], *args[4]);
}

static void
save_prefs(user_specs_t *specs, char *config_path, GtkWidget *window, ui_prefs_elements_t *ui_prefs_elements, GtkWidget *main_window)
{
	user_specs_t save_specs;

	init_user_specs(&save_specs);

	save_specs.main_window_always_on_top = specs->main_window_always_on_top = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (ui_prefs_elements->main_window_on_top_box));
	save_specs.progress_window_always_on_top = specs->progress_window_always_on_top = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (ui_prefs_elements->progress_window_on_top_box));
	save_specs.show_tray = specs->show_tray = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (ui_prefs_elements->show_tray_box));

	gtk_window_set_keep_above(GTK_WINDOW(main_window), save_specs.main_window_always_on_top);

	if (!config_path)
		show_stderr("No configuration file open, preferences will not be remembered");
	else if (!load(&save_specs, config_path) && errno != ENOENT)	/* If there ain't an existing configuration file, we'll just write a new 'un */
		show_warning_dialog("Couldn't load configuration file \"%s\". Error: \"%s\" (%d)", config_path, strerror(errno), errno);
	else
	{
		/* Note that save_specs.auto_run will already be set if set in the config_file, after the load() above */

		if (save(&save_specs, config_path) == FALSE)
			show_warning_dialog("Unable to save preferences to configuration file \"%s\". Preferences not saved.\n\nError \"%s\" (%d)", config_path, strerror(errno), errno);
	}

	gtk_widget_destroy(window);
}

static void
select_and_open_config_cb(void ***args)
{
	select_and_open_config(*args[0], *args[1], *args[2], *args[3], *args[4]);
}

static void
select_and_open_config(user_specs_t *specs, char **config_path, GtkWidget *main_window, ui_specs_elements_t *ui_specs_elements, GtkWidget *save_menu_item)
{
	GtkWidget *file_selection;

	file_selection = gtk_file_chooser_dialog_new("Select configuration file", GTK_WINDOW (main_window), GTK_FILE_CHOOSER_ACTION_OPEN, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

	{
		GtkFileFilter *filter;

		filter = gtk_file_filter_new();
		gtk_file_filter_add_pattern(filter, "*." CONFIG_EXTENSION);
		gtk_file_filter_add_mime_type(filter, CONFIG_MIME_TYPE);
		gtk_file_filter_set_name(filter, PRGNAME " configuration");
		gtk_file_chooser_set_filter(GTK_FILE_CHOOSER (file_selection), filter);
	}

	if (gtk_dialog_run(GTK_DIALOG (file_selection)) == GTK_RESPONSE_ACCEPT)
	{
		char *new_config_path;

		if (!(new_config_path = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER (file_selection))))	/* "if no file is selected, or the selected file can't be represented with a local filename" - http://library.gnome.org/devel/gtk/stable/GtkFileChooser.html#gtk-file-chooser-get-filename */
		{
			gtk_widget_destroy(file_selection);

			return;
		}

		/* Rewrite the config path so that we can free() the string, rather than g_free() it */
		{
			char *temp;

			if ((pasprintf(&temp, "%s", new_config_path)) == -1)
				show_error_dialog(EX_OSERR, "pasprintf() returned -1 in %s(). Error: \"%s\" (%d)", __func__, strerror(errno), errno);

			g_free(new_config_path);

			new_config_path = temp;
		}

		if (load(specs, new_config_path))
		{
			*config_path = new_config_path;

			update_main(specs, *config_path, ui_specs_elements, save_menu_item);
		}
		else
			show_warning_dialog("Unable to open configuration file \"%s\". Error: \"%s\" (%d)", new_config_path, strerror(errno), errno);
	}

	gtk_widget_destroy(file_selection);
}

static void
show_about(void)
{
	GtkWidget *about_dialog;
	char *authors[] = AUTHORS;

	about_dialog = gtk_about_dialog_new();

	gtk_about_dialog_set_name(GTK_ABOUT_DIALOG (about_dialog), PRGNAME);
	gtk_about_dialog_set_version(GTK_ABOUT_DIALOG (about_dialog), VERSION);
	gtk_about_dialog_set_copyright(GTK_ABOUT_DIALOG (about_dialog), COPYRIGHT);
	gtk_about_dialog_set_website(GTK_ABOUT_DIALOG (about_dialog), CONTACT_WWW);
	gtk_about_dialog_set_license(GTK_ABOUT_DIALOG (about_dialog), LICENSE_LONG);
	gtk_about_dialog_set_authors(GTK_ABOUT_DIALOG (about_dialog), (const gchar **) authors);

	g_signal_connect(GTK_ABOUT_DIALOG (about_dialog), "response", G_CALLBACK (gtk_widget_destroy), NULL);

	gtk_widget_show(about_dialog);
}

void
show_error_dialog(int eval, const char *fmt, ...)
{
	va_list ap;
	char *exp_fmt;
	GtkWidget *message_dialog;

	va_start(ap, fmt);

	if ((pvasprintf(&exp_fmt, fmt, ap)) == -1)
	{
		show_stderr("pvasprintf() returned -1 in %s(). Error: \"%s\" (%d)", __func__, strerror(errno), errno);

		exit(EX_OSERR);
	}

	va_end(ap);

	show_stderr(exp_fmt);

	message_dialog = gtk_message_dialog_new(NULL, (GtkDialogFlags) 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE, exp_fmt);
	gtk_window_set_title(GTK_WINDOW (message_dialog), "Error - " PRGNAME);
	gtk_dialog_run(GTK_DIALOG (message_dialog));
	gtk_widget_destroy(message_dialog);

	exit(eval);
}

void
show_main_window(user_specs_t *specs, char **config_path)
{
	GtkAccelGroup *accel_group;
	GtkWidget *window, *save_menu_item, *menu_bar, *menu, *menu_item, *vbox, *hbox, *label, *separator, *button, *ok_button;
	ui_specs_elements_t *ui_specs_elements;
	void **select_and_open_config_cb_args[5], **save_current_cb_args[3], **save_as_cb_args[4], **show_prefs_cb_args[3], **ui_start_shutdown_cb_args[3];

	select_and_open_config_cb_args[0] = (void **) &specs;
	select_and_open_config_cb_args[1] = (void **) &config_path;
	select_and_open_config_cb_args[2] = (void **) &window;
	select_and_open_config_cb_args[3] = (void **) &ui_specs_elements;
	select_and_open_config_cb_args[4] = (void **) &save_menu_item;
	save_current_cb_args[0] = (void **) &specs;
	save_current_cb_args[1] = (void **) config_path;
	save_current_cb_args[2] = (void **) &ui_specs_elements;
	save_as_cb_args[0] = (void **) &config_path;
	save_as_cb_args[1] = (void **) &specs;
	save_as_cb_args[2] = (void **) &ui_specs_elements;
	save_as_cb_args[3] = (void **) &window;
	show_prefs_cb_args[0] = (void **) &specs;
	show_prefs_cb_args[1] = (void **) config_path;
	show_prefs_cb_args[2] = (void **) &window;
	ui_start_shutdown_cb_args[0] = (void **) &specs;
	ui_start_shutdown_cb_args[1] = (void **) &window;
	ui_start_shutdown_cb_args[2] = (void **) &ui_specs_elements;

	if ((ui_specs_elements = malloc(sizeof (*ui_specs_elements))) == NULL)
		show_error_dialog(EX_OSERR, "malloc() returned NULL in %s(). Error: \"%s\" (%d)", __func__, strerror(errno), errno);

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	g_signal_connect(G_OBJECT (window), "delete_event", G_CALLBACK (gtk_widget_destroy), NULL);
	g_signal_connect(G_OBJECT (window), "destroy", G_CALLBACK (gtk_main_quit), NULL);
	gtk_window_set_title(GTK_WINDOW (window), PRGNAME);
	gtk_window_set_resizable(GTK_WINDOW (window), FALSE);
	gtk_window_set_keep_above(GTK_WINDOW (window), specs->main_window_always_on_top);

	accel_group = gtk_accel_group_new();
	gtk_window_add_accel_group(GTK_WINDOW (window), accel_group);

	vbox = gtk_vbox_new(FALSE, 0);

	menu_bar = gtk_menu_bar_new();
	gtk_box_pack_start(GTK_BOX (vbox), menu_bar, FALSE, FALSE, 0);

	/* File menu */
	menu = gtk_menu_new();

	menu_item = gtk_image_menu_item_new_from_stock(GTK_STOCK_OPEN, accel_group);
	g_signal_connect_swapped(G_OBJECT (menu_item), "activate", G_CALLBACK (select_and_open_config_cb), select_and_open_config_cb_args);
	gtk_menu_shell_append(GTK_MENU_SHELL (menu), menu_item);
	gtk_widget_show(menu_item);

	save_menu_item = gtk_image_menu_item_new_from_stock(GTK_STOCK_SAVE, accel_group);
	gtk_widget_set_sensitive(save_menu_item, (*config_path != NULL));
	g_signal_connect_swapped(G_OBJECT (save_menu_item), "activate", G_CALLBACK (save_current_cb), save_current_cb_args);
	gtk_menu_shell_append(GTK_MENU_SHELL (menu), save_menu_item);
	gtk_widget_show(save_menu_item);

	menu_item = gtk_image_menu_item_new_from_stock(GTK_STOCK_SAVE_AS, accel_group);
	g_signal_connect_swapped(G_OBJECT (menu_item), "activate", G_CALLBACK (save_as_cb), save_as_cb_args);
	gtk_menu_shell_append(GTK_MENU_SHELL (menu), menu_item);
	gtk_widget_show(menu_item);

	menu_item = gtk_separator_menu_item_new();
	gtk_menu_shell_append(GTK_MENU_SHELL (menu), menu_item);
	gtk_widget_show(menu_item);

	menu_item = gtk_image_menu_item_new_from_stock(GTK_STOCK_QUIT, accel_group);
	g_signal_connect(G_OBJECT (menu_item), "activate", G_CALLBACK (gtk_main_quit), NULL);
	gtk_menu_shell_append(GTK_MENU_SHELL (menu), menu_item);
	gtk_widget_show(menu_item);

	menu_item = gtk_menu_item_new_with_mnemonic("_File");
	gtk_widget_show(menu_item);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM (menu_item), menu);
	gtk_menu_bar_append(GTK_MENU_BAR (menu_bar), menu_item);

	/* Edit menu */
	menu = gtk_menu_new();

	menu_item = gtk_image_menu_item_new_from_stock(GTK_STOCK_PREFERENCES, accel_group);
	g_signal_connect_swapped(G_OBJECT (menu_item), "activate", G_CALLBACK (show_prefs_cb), show_prefs_cb_args);
	gtk_menu_shell_append(GTK_MENU_SHELL (menu), menu_item);
	gtk_widget_show(menu_item);

	menu_item = gtk_menu_item_new_with_mnemonic("_Edit");
	gtk_widget_show(menu_item);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM (menu_item), menu);
	gtk_menu_bar_append(GTK_MENU_BAR (menu_bar), menu_item);

	/* Help menu */
	menu = gtk_menu_new();

	menu_item = gtk_image_menu_item_new_from_stock(GTK_STOCK_ABOUT, accel_group);
	g_signal_connect_swapped(G_OBJECT (menu_item), "activate", G_CALLBACK (show_about), NULL);
	gtk_menu_shell_append(GTK_MENU_SHELL (menu), menu_item);
	gtk_widget_show(menu_item);

	menu_item = gtk_menu_item_new_with_mnemonic("_Help");
	gtk_widget_show(menu_item);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM (menu_item), menu);
	gtk_menu_bar_append(GTK_MENU_BAR (menu_bar), menu_item);

	gtk_widget_show(menu_bar);

	/* Timeout selection */
	hbox = gtk_hbox_new(FALSE, 0);

	label = gtk_label_new("Timeout:");
	gtk_box_pack_start(GTK_BOX (hbox), label, FALSE, FALSE, 5);
	gtk_widget_show(label);

	ui_specs_elements->timeout_field = gtk_entry_new();
	g_signal_connect_swapped(G_OBJECT (ui_specs_elements->timeout_field), "activate", G_CALLBACK (ui_start_shutdown_cb), ui_start_shutdown_cb_args);
	gtk_box_pack_start(GTK_BOX (hbox), ui_specs_elements->timeout_field, FALSE, FALSE, 5);
	gtk_widget_show(ui_specs_elements->timeout_field);

	ui_specs_elements->timeout_unit_combo_box = gtk_combo_box_new_text();
	gtk_combo_box_append_text(GTK_COMBO_BOX (ui_specs_elements->timeout_unit_combo_box), "minutes");	/* Be careful of order here - replicate in TIMEOUT_UNITS (in downtime.h) */
	gtk_combo_box_append_text(GTK_COMBO_BOX (ui_specs_elements->timeout_unit_combo_box), "hours");
	gtk_combo_box_append_text(GTK_COMBO_BOX (ui_specs_elements->timeout_unit_combo_box), "days");
	gtk_combo_box_set_active(GTK_COMBO_BOX (ui_specs_elements->timeout_unit_combo_box), (gint) specs->timeout_unit);
	gtk_box_pack_start(GTK_BOX (hbox), ui_specs_elements->timeout_unit_combo_box, FALSE, FALSE, 5);
	gtk_widget_show(ui_specs_elements->timeout_unit_combo_box);

	gtk_box_pack_start(GTK_BOX (vbox), hbox, FALSE, TRUE, 5);
	gtk_widget_show(hbox);

	/* Shutdown type selection */
	hbox = gtk_hbox_new(FALSE, 0);

	label = gtk_label_new("Shutdown type:");
	gtk_box_pack_start(GTK_BOX (hbox), label, FALSE, FALSE, 5);
	gtk_widget_show(label);

	ui_specs_elements->shutdown_type_combo_box = gtk_combo_box_new_text();
	gtk_combo_box_append_text(GTK_COMBO_BOX (ui_specs_elements->shutdown_type_combo_box), "Power-off");	/* Be careful of order here - replicate in SHUTDOWN_TYPES (in downtime.h) */
	gtk_combo_box_append_text(GTK_COMBO_BOX (ui_specs_elements->shutdown_type_combo_box), "Restart");
	gtk_combo_box_append_text(GTK_COMBO_BOX (ui_specs_elements->shutdown_type_combo_box), "Halt");
	gtk_combo_box_append_text(GTK_COMBO_BOX (ui_specs_elements->shutdown_type_combo_box), "Single user mode");
	gtk_combo_box_set_active(GTK_COMBO_BOX (ui_specs_elements->shutdown_type_combo_box), (gint) specs->shutdown_type);
	gtk_box_pack_start(GTK_BOX (hbox), ui_specs_elements->shutdown_type_combo_box, FALSE, TRUE, 5);
	gtk_widget_show(ui_specs_elements->shutdown_type_combo_box);

	gtk_box_pack_start(GTK_BOX (vbox), hbox, FALSE, TRUE, 5);
	gtk_widget_show(hbox);

	separator = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX (vbox), separator, FALSE, TRUE, 5);
	gtk_widget_show(separator);

	/*
	* OK / Cancel buttons
	*/

	hbox = gtk_hbox_new(FALSE, 0);

	button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
	g_signal_connect_swapped(G_OBJECT (button), "clicked", G_CALLBACK (gtk_main_quit), NULL);
	gtk_box_pack_start(GTK_BOX (hbox), button, FALSE, TRUE, 5);
	gtk_widget_show(button);

	ok_button = gtk_button_new_from_stock(GTK_STOCK_OK);
	g_signal_connect_swapped(G_OBJECT (ok_button), "clicked", G_CALLBACK (ui_start_shutdown_cb), ui_start_shutdown_cb_args);
	gtk_box_pack_start(GTK_BOX (hbox), ok_button, FALSE, TRUE, 5);
	gtk_widget_show(ok_button);

	gtk_box_pack_start(GTK_BOX (vbox), hbox, FALSE, TRUE, 5);
	gtk_widget_show(hbox);

	gtk_container_add(GTK_CONTAINER (window), vbox);
	gtk_widget_show(vbox);

	GTK_WIDGET_SET_FLAGS(ok_button, GTK_CAN_DEFAULT);
	gtk_widget_grab_default(ok_button);

	update_main(specs, *config_path, ui_specs_elements, save_menu_item);

	gtk_widget_show(GTK_WIDGET (window));

	gtk_main();

	free(ui_specs_elements);
}

static void
show_prefs_cb(void ***args)
{
	show_prefs(*args[0], *args[1], *args[2]);
}

static void
show_prefs(user_specs_t *specs, char *config_path, GtkWidget *main_window)
{
	GtkWidget *window, *vbox, *hbox, *button_cancel, *button_ok;
	ui_prefs_elements_t *ui_prefs_elements;
	void **save_prefs_cb_args[5];

	save_prefs_cb_args[0] = (void **) &specs;
	save_prefs_cb_args[1] = (void **) &config_path;
	save_prefs_cb_args[2] = (void **) &window;
	save_prefs_cb_args[3] = (void **) &ui_prefs_elements;
	save_prefs_cb_args[4] = (void **) &main_window;

	if ((ui_prefs_elements = malloc(sizeof (*ui_prefs_elements))) == NULL)
		show_error_dialog(EX_OSERR, "malloc() returned NULL in %s(). Error: \"%s\" (%d)", __func__, strerror(errno), errno);

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), "Preferences - " PRGNAME);
	gtk_window_set_resizable(GTK_WINDOW(window), FALSE);
	gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);

	vbox = gtk_vbox_new(FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	ui_prefs_elements->main_window_on_top_box = gtk_check_button_new_with_mnemonic("Main window always on _top");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (ui_prefs_elements->main_window_on_top_box), specs->main_window_always_on_top);
	gtk_box_pack_start(GTK_BOX (hbox), ui_prefs_elements->main_window_on_top_box, FALSE, TRUE, 5);
	gtk_widget_show(ui_prefs_elements->main_window_on_top_box);
	gtk_box_pack_start(GTK_BOX (vbox), hbox, FALSE, TRUE, 5);
	gtk_widget_show(hbox);

	hbox = gtk_hbox_new(FALSE, 0);
	ui_prefs_elements->progress_window_on_top_box = gtk_check_button_new_with_mnemonic("_Progress window always on top");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (ui_prefs_elements->progress_window_on_top_box), specs->progress_window_always_on_top);
	gtk_box_pack_start(GTK_BOX (hbox), ui_prefs_elements->progress_window_on_top_box, FALSE, TRUE, 5);
	gtk_widget_show(ui_prefs_elements->progress_window_on_top_box);
	gtk_box_pack_start(GTK_BOX (vbox), hbox, FALSE, TRUE, 5);
	gtk_widget_show(hbox);

	hbox = gtk_hbox_new(FALSE, 0);
	ui_prefs_elements->show_tray_box = gtk_check_button_new_with_mnemonic("_Show tray icon during shutdown");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (ui_prefs_elements->show_tray_box), specs->show_tray);
	gtk_box_pack_start(GTK_BOX (hbox), ui_prefs_elements->show_tray_box, FALSE, TRUE, 5);
	gtk_widget_show(ui_prefs_elements->show_tray_box);
	gtk_box_pack_start(GTK_BOX (vbox), hbox, FALSE, TRUE, 5);
	gtk_widget_show(hbox);

	hbox = gtk_hbox_new(FALSE, 0);
	button_cancel = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
	button_ok = gtk_button_new_from_stock(GTK_STOCK_OK);
	gtk_box_pack_start(GTK_BOX (hbox), button_cancel, FALSE, TRUE, 5);
	gtk_box_pack_start(GTK_BOX (hbox), button_ok, FALSE, TRUE, 5);
	gtk_widget_show(button_ok);
	gtk_widget_show(button_cancel);
	gtk_box_pack_start(GTK_BOX (vbox), hbox, FALSE, TRUE, 5);
	gtk_widget_show(hbox);

	gtk_container_add(GTK_CONTAINER (window), vbox);
	gtk_widget_show(vbox);

	g_signal_connect(G_OBJECT (window), "delete_event", G_CALLBACK (gtk_widget_destroy), NULL);
	g_signal_connect(G_OBJECT (window), "destroy", G_CALLBACK (gtk_main_quit), NULL);
	g_signal_connect_swapped(G_OBJECT (button_cancel), "clicked", G_CALLBACK (gtk_widget_destroy), window);
	g_signal_connect_swapped(G_OBJECT (button_ok), "clicked", G_CALLBACK (save_prefs_cb), save_prefs_cb_args);

	GTK_WIDGET_SET_FLAGS(button_ok, GTK_CAN_DEFAULT);
	gtk_widget_grab_default(button_ok);

	gtk_widget_show(window);

	gtk_main();

	free(ui_prefs_elements);
}

void
show_progress_window(shutdown_t *shutdown, user_specs_t *specs)
{
	GtkWidget *window, *progress_bar, *vbox, *hbox, *pause_button, *close_button;
	GtkStatusIcon *tray_icon;
	char *str_time_to_shutdown;
	void **destroy_progress_cb_args[2], **update_progress_cb_args[4], **pause_shutdown_cb_args[3];

	destroy_progress_cb_args[0] = (void **) &shutdown;
	destroy_progress_cb_args[1] = (void **) &window;
	update_progress_cb_args[0] = (void **) &shutdown;
	update_progress_cb_args[1] = (void **) &progress_bar;
	update_progress_cb_args[2] = (void **) &tray_icon;
	update_progress_cb_args[3] = (void **) &pause_button;
	pause_shutdown_cb_args[0] = (void **) &shutdown;
	pause_shutdown_cb_args[1] = (void **) &progress_bar;
	pause_shutdown_cb_args[2] = (void **) &tray_icon;

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	g_signal_connect_swapped(G_OBJECT (window), "delete_event", G_CALLBACK (destroy_progress_cb), destroy_progress_cb_args);
	g_signal_connect_swapped(G_OBJECT (window), "destroy", G_CALLBACK (destroy_progress_cb), destroy_progress_cb_args);

	gtk_window_set_title(GTK_WINDOW (window), "Shutdown progress - " PRGNAME);
	gtk_window_set_resizable(GTK_WINDOW (window), FALSE);
	gtk_window_set_position(GTK_WINDOW (window), GTK_WIN_POS_CENTER);
	gtk_window_set_keep_above(GTK_WINDOW(window), specs->progress_window_always_on_top);
	gtk_container_set_border_width (GTK_CONTAINER (window), 10);

	vbox = gtk_vbox_new(FALSE, 0);

	progress_bar = gtk_progress_bar_new();
	gtk_progress_bar_set_text(GTK_PROGRESS_BAR (progress_bar), (str_time_to_shutdown = get_shutdown_time_left_string(shutdown)));
	g_timeout_add(PROGRESS_UPDATE_PERIOD, (GSourceFunc) update_progress_cb, update_progress_cb_args);

	free(str_time_to_shutdown);
	gtk_box_pack_start(GTK_BOX (vbox), progress_bar, FALSE, FALSE, 5);
	gtk_widget_show(progress_bar);

	hbox = gtk_hbox_new(FALSE, 0);

	pause_button = gtk_button_new_from_stock(GTK_STOCK_MEDIA_PAUSE);
	g_signal_connect_swapped(G_OBJECT (pause_button), "clicked", G_CALLBACK (pause_shutdown_cb), pause_shutdown_cb_args);

	if (!shutdown->pid)
		gtk_widget_set_sensitive(pause_button, FALSE);	/* Disable the pause button if we don't know the shutdown(8) process (if any)' PID */

	gtk_box_pack_start(GTK_BOX (hbox), pause_button, FALSE, FALSE, 5);
	gtk_widget_show(pause_button);

	close_button = gtk_button_new_from_stock(GTK_STOCK_CLOSE);
	g_signal_connect_swapped(G_OBJECT (close_button), "clicked", G_CALLBACK (destroy_progress_cb), destroy_progress_cb_args);
	gtk_box_pack_start(GTK_BOX (hbox), close_button, FALSE, FALSE, 5);
	gtk_widget_show(close_button);

	gtk_box_pack_start(GTK_BOX (vbox), hbox, FALSE, FALSE, 5);
	gtk_widget_show(hbox);
	gtk_container_add(GTK_CONTAINER (window), vbox);

	if (specs->show_tray)
		tray_icon = show_tray_icon(window, destroy_progress_cb_args);
	else 
		tray_icon = NULL;

	gtk_widget_show(vbox);
	gtk_widget_show(window);

	gtk_main();
}

void
show_stderr(const char *fmt, ...)
{
	va_list ap;
	char *exp_fmt;
	time_t curr_time;
	struct tm *tm;

	va_start(ap, fmt);

	if ((pvasprintf(&exp_fmt, fmt, ap)) == -1)
		show_error_dialog(EX_OSERR, "pvasprintf() returned -1 in %s(). Error: \"%s\" (%d)", __func__, strerror(errno), errno);

	va_end(ap);

	curr_time = time((time_t *) NULL);
	tm = localtime(&curr_time);

	fprintf(stderr, "%d-%.2d-%.2d %.2d:%.2d:%.2d %s: %s\n", tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec, PRGNAME, exp_fmt);
}

static GtkStatusIcon *
show_tray_icon(GtkWidget *show_hide_window, void ***destroy_progress_cb_args)
{
	GtkStatusIcon *tray_icon;
	GtkWidget *menu, *menu_item;
	void **tray_icon_click_cb_args[1];

	tray_icon_click_cb_args[0] = (void **) &show_hide_window;

	tray_icon = gtk_status_icon_new();

	menu = gtk_menu_new();

	g_signal_connect_swapped(G_OBJECT (tray_icon), "activate", G_CALLBACK (tray_icon_click_cb), tray_icon_click_cb_args);
	g_signal_connect(G_OBJECT (tray_icon), "popup-menu", G_CALLBACK (tray_icon_menu), menu);

	menu_item = gtk_menu_item_new_with_label ("Show / hide progress");
	g_signal_connect_swapped(G_OBJECT (menu_item), "activate", G_CALLBACK (tray_icon_click_cb), tray_icon_click_cb_args);
	gtk_menu_shell_append(GTK_MENU_SHELL (menu), menu_item);

	menu_item = gtk_menu_item_new_with_label ("Close");
	g_signal_connect_swapped(G_OBJECT (menu_item), "activate", G_CALLBACK (destroy_progress_cb), destroy_progress_cb_args);
	gtk_menu_shell_append(GTK_MENU_SHELL (menu), menu_item);

	gtk_widget_show_all(menu);

	gtk_status_icon_set_from_icon_name(tray_icon, GTK_STOCK_DISCONNECT);
	gtk_status_icon_set_title(tray_icon, PRGNAME);
	gtk_status_icon_set_tooltip_text(tray_icon, PRGNAME);
	gtk_status_icon_set_visible(tray_icon, TRUE);

	return (tray_icon);
}

void
show_warning_dialog(const char *fmt, ...)
{
	va_list ap;
	char *exp_fmt;
	GtkWidget *message_dialog;

	va_start(ap, fmt);

	if ((pvasprintf(&exp_fmt, fmt, ap)) == -1)
		show_error_dialog(EX_OSERR, "pvasprintf() returned -1 in %s(). Error: \"%s\" (%d)", __func__, strerror(errno), errno);

	va_end(ap);

	show_stderr(exp_fmt);

	message_dialog = gtk_message_dialog_new(NULL, (GtkDialogFlags) 0, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK, exp_fmt);
	gtk_window_set_title(GTK_WINDOW (message_dialog), "Warning - " PRGNAME);
	gtk_dialog_run(GTK_DIALOG (message_dialog));
	gtk_widget_destroy(message_dialog);

	free(exp_fmt);
}

static void
tray_icon_click_cb(void ***args)
{
	tray_icon_click(*args[0]);
}

static void
tray_icon_click(GtkWidget *show_hide_window)
{
	if (show_hide_window)
	{
		if (gtk_widget_get_visible(GTK_WIDGET (show_hide_window)))
			gtk_widget_hide(GTK_WIDGET (show_hide_window));
		else
			gtk_widget_show(GTK_WIDGET (show_hide_window));
	}
}

static void
tray_icon_menu(GtkStatusIcon *tray_icon, guint button, guint32 activate_time, gpointer menu)
{
	gtk_menu_popup(GTK_MENU (menu), NULL, NULL, gtk_status_icon_position_menu, tray_icon, button, activate_time);
}

static shutdown_t *
ui_start_shutdown_cb(void ***args)
{
	return (ui_start_shutdown(*args[0], *args[1], *args[2]));
}

static shutdown_t *
ui_start_shutdown(user_specs_t *specs, GtkWidget *main_window, ui_specs_elements_t *ui_specs_elements)
{
	extern char *base_config_path;

	shutdown_t *shutdown;

	if ((shutdown = malloc(sizeof (*shutdown))) == NULL)
		show_error_dialog(EX_OSERR, "malloc() returned NULL in %s(). Error: \"%s\" (%d)", __func__, strerror(errno), errno);

	if (update_user_specs(specs, ui_specs_elements) == NULL || parse_specs(specs, shutdown) == NULL)	/* If we weren't able to successfully read the user specs... Note also that update_user_specs() deals with alerting the user about input dramas */
		return (NULL);

	specs->auto_run = FALSE;	/* Never write auto_run to the user default configuration file */

	if (save(specs, base_config_path) == FALSE)
		show_stderr("Unable to save to configuration file \"%s\". Settings will not be saved. \n\nError \"%s\" (%d)", base_config_path, strerror(errno), errno);

	if (start_shutdown(shutdown))
	{
		gtk_widget_destroy(GTK_WIDGET (main_window));

		show_progress_window(shutdown, specs);
	}

	return (shutdown);
}

static void
update_main(user_specs_t *specs, char *config_path, ui_specs_elements_t *ui_specs_elements, GtkWidget *save_menu_item)
{
	char *timeout;

	if ((pasprintf(&timeout, "%lu", specs->timeout)) == -1)
		show_error_dialog(EX_OSERR, "pasprintf() returned -1 in %s(). Error: \"%s\" (%d)", __func__, strerror(errno), errno);

	gtk_entry_set_text(GTK_ENTRY (ui_specs_elements->timeout_field), timeout);

	free(timeout);

	gtk_combo_box_set_active(GTK_COMBO_BOX (ui_specs_elements->timeout_unit_combo_box), (gint) specs->timeout_unit);
	gtk_combo_box_set_active(GTK_COMBO_BOX (ui_specs_elements->shutdown_type_combo_box), (gint) specs->shutdown_type);
	gtk_widget_set_sensitive(save_menu_item, (config_path != NULL));
}

static gboolean
update_progress_cb(void ***args)
{
	return (update_progress(*args[0], *args[1], *args[2], *args[3]));
}

static gboolean
update_progress(shutdown_t *shutdown, GtkWidget *progress_bar, GtkStatusIcon *tray_icon, GtkWidget *pause_button)
{
	int status;

	if (shutdown->paused)	/* If we're paused, we don't need to do anything. Rather than cancel the callback (and reinitialising it when resuming), we just do nothing */
		return (TRUE);

	if (difftime(shutdown->shutdown_time, time(NULL)) <= 0)	/* If the shutdown time has elapsed */
	{
		gtk_widget_set_sensitive(pause_button, FALSE);	/* Disable pause button - it can't help now! */

		gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR (progress_bar), 1);	/* 100% done */
		gtk_progress_bar_set_text(GTK_PROGRESS_BAR (progress_bar), "(Shutting down)");

		if (tray_icon)
			gtk_status_icon_set_tooltip_text(tray_icon, "DownTime - (Shutting down)");

		return (FALSE);	/* We won't update any further */
	}

	waitpid(shutdown->pid, &status, WNOHANG);

	if (!no_check_proc && shutdown->pid && kill(shutdown->pid, 0) != 0 && WIFEXITED (status))	/* Ensure that shutdown(8) is still running. shutdown->pid is checked to ensure it's > 0 in shutdown.c */
	{
		if (errno == EPERM)	/* From kill() */
		{
			show_warning_dialog("shutdown(8) (PID: %ld) status could not be checked due to lack of permissions.\n\nPlease ensure that shutdown(8) is running correctly", shutdown->pid);

			no_check_proc = TRUE;
		}
		else
		{
			if (WIFEXITED (status))
				show_error_dialog(EX_UNAVAILABLE, "shutdown(8) (PID: %ld) exited unexpectedly with exit status %d (\"%s\")", shutdown->pid, WEXITSTATUS (status), strerror(WEXITSTATUS (status)));
			else
				show_error_dialog(EX_UNAVAILABLE, "shutdown(8) (PID: %ld) terminated unexpectedly", shutdown->pid);
		}
	}

	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR (progress_bar), progress_fraction(shutdown));

	{
		char *str_time_to_shutdown;

		gtk_progress_bar_set_text(GTK_PROGRESS_BAR (progress_bar), (str_time_to_shutdown = get_shutdown_time_left_string(shutdown)));

		if (tray_icon)
			gtk_status_icon_set_tooltip_text(tray_icon, str_time_to_shutdown);

		free(str_time_to_shutdown);
	}

	return (TRUE);
}

/* Updates the structure pointed to by specs with user-supplied specifications from the DownTime main dialogue. DO NOT call after the main window has been destroyed */
static user_specs_t *
update_user_specs(user_specs_t *specs, ui_specs_elements_t *ui_specs_elements)
{
	char *nptr, *endptr;
	unsigned long int timeout;

	nptr = (char *) gtk_entry_get_text(GTK_ENTRY (ui_specs_elements->timeout_field));
	timeout = (unsigned long int) strtol(nptr, &endptr, 10);

	if (!(*nptr != '\0' && *endptr == '\0'))
	{
		show_warning_dialog("An invalid timeout - \"%s\" - was specified", nptr);
	
		return (NULL);
	}

	specs->timeout = timeout;
	specs->timeout_unit = gtk_combo_box_get_active(GTK_COMBO_BOX (ui_specs_elements->timeout_unit_combo_box));
	specs->shutdown_type = gtk_combo_box_get_active(GTK_COMBO_BOX (ui_specs_elements->shutdown_type_combo_box));

	return (specs);
}

