/*-
* Copyright (C) 2005-2010 Nick Withers. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) AND CONTRIBUTORS ``AS IS'' AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR(S) OR CONTRIBUTORS BE LIABLE
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
* OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
* LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
* OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGE.
*/

#include "downtime.h"

#include <errno.h>
#include <gtk/gtk.h>
#include <pwd.h>
#include <stdio.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

#ifdef BSD
#include <sysexits.h>
#endif

char *shutdown_path;
char *base_config_path;

int
main(int argc, char *argv[])
{
	user_specs_t specs;
	char launched_with_config, *config_path;

	gtk_init(&argc, &argv);	/* Parse GTK command-line parameters */
	g_set_prgname(PRGNAME);	/* Set the program name */

	init_user_specs(&specs);

	/* Allow for the specification of a configuration file. Done a little primitively... */
	if (argc == 2)
	{
		config_path = argv[1];

		launched_with_config = TRUE;
	}
	else
	{
		char found_user;

		if ((config_path = base_config_path = get_config_path(&found_user)) == NULL)
			show_warning_dialog("%s Default settings will be used and settings will not be remembered", (found_user ? "Couldn't locate the default configuration file location." : "Couldn't find the user's passwd entry."));
	}

	if (config_path && !load(&specs, config_path))	/* If we couldn't load configuration information from the file at config_path */
	{
		if (errno == ENOENT)
			show_stderr("Couldn't find configuration file \"%s\".\n\nDefault shutdown parameters will be used", config_path);	/* If we couldn't read the configuration file because no such file exists, don't warn the user graphically, just print a brief message to stdout */
		else
			show_warning_dialog("Couldn't parse configuration file \"%s\".\n\nError: \"%s\" (%d)", config_path, strerror(errno), errno);

		config_path = NULL;
	}

#ifndef WITH_SHUTDOWN
	if ((shutdown_path = find_executable_in_path(SHUTDOWN_FILENAME)) == NULL)
		shutdown_path = DEFAULT_SHUTDOWN_PATH;
#else
	shutdown_path = WITH_SHUTDOWN;
#endif

	if (access(shutdown_path, F_OK) != 0)
		show_error_dialog(EX_UNAVAILABLE, PRGNAME " was unable to find the shutdown binary, required for operation.\n\nThe search path was checked and the /sbin/shutdown binary does not exist");

	{
		struct stat sb;

		if (stat(shutdown_path, &sb) != 0)
			show_warning_dialog("Couldn't stat(2) \"%s\". Error: \"%s\" (%d)", shutdown_path, strerror(errno), errno);
		else if (!((sb.st_mode & S_ISUID) && sb.st_uid == 0) && geteuid() != 0)	/* If the shutdown(8) binary isn't setuid root and we don't have superuser privileges */
			show_warning_dialog("The shutdown(8) binary \"%s\" is not setuid root and you don't have superuser permissions.\n\nYou may have to launch " PRGNAME " as the superuser or modify the binary's permissions to be able to successfully perform a shutdown", shutdown_path);
		else
		{
			if (access(shutdown_path, X_OK) != 0)
				show_warning_dialog("You do not appear to have \"execute\" permissions on the shutdown(8) binary and may not be able to successfully perform a shutdown.\n\nRunning \"chmod o+x %s\" may resolve this, but please be aware of the security ramifications of doing so", shutdown_path);
		}
	}

	if (launched_with_config && specs.auto_run)
	{
		shutdown_t *shutdown;

		if ((shutdown = malloc(sizeof (*shutdown))) == NULL)
			show_error_dialog(EX_OSERR, "malloc() returned NULL in %s(). Error: \"%s\" (%d)", __func__, strerror(errno), errno);

		if (parse_specs(&specs, shutdown) == NULL)
			show_error_dialog(EX_OSERR, "Couldn't parse configuration file \"%s\"", config_path);

		if (start_shutdown(shutdown))
			show_progress_window(shutdown, &specs);
	}
	else
		show_main_window(&specs, &config_path);

	return (EXIT_SUCCESS);
}
